from asyncio import gather


#
# I like to think the common use-case for 'gather' is to cancel all async
#   operations should any one of them fail.  This is not how 'gather' was
#   implemented though which is why I wrote this function
#
async def resolveAll(awaitables):
    gatherAll = None
    try:
        gatherAll = gather(*awaitables)
        result = await gatherAll
    finally:
        if gatherAll:
            gatherAll.cancel()

    return result


def iif(condition, whenTruthy, whenFalsey):
    if condition:
        return whenTruthy
    else:
        return whenFalsey



#
# TODO: find a cross-platform way to rely only on the returncode for untracked
#   files.  Currently it returns stdout/stderr which should be unnecessary.
#

# ------- #
# Imports #
# ------- #

from textwrap import dedent
from .utils import iif, resolveAll
import asyncio
import os
import subprocess


# ---- #
# Init #
# ---- #

runAsync = asyncio.create_subprocess_exec
run = subprocess.run

headExistsCmd = ["git", "rev-parse", "HEAD"]
isAGitRepoCmd = ["git", "rev-parse", "--is-inside-work-tree"]
refreshIndexCmd = ["git", "update-index", "--refresh"]
# Any way to remove the double negative?
allFilesAreTrackedCmd = [
    "git",
    "ls-files",
    "--other",
    "--directory",
    "--exclude-standard",
    "--no-empty-directory",
]
emptyGitTree = "4b825dc642cb6eb9a060e54bf8d69288fbee4904"


# ---- #
# Main #
# ---- #


async def check(cwd=None):
    cwd = cleanCwd(cwd)

    isAGitRepo_proc = await runAsync(
        *isAGitRepoCmd,
        stdout=asyncio.subprocess.DEVNULL,
        stderr=asyncio.subprocess.DEVNULL,
        cwd=cwd,
    )
    isAGitRepo_result = await isAGitRepo_proc.wait()
    if isAGitRepo_result != 0:
        raise Exception(f"'{cwd}' is not a git repository")

    headExists_proc = await runAsync(
        *headExistsCmd,
        stdout=asyncio.subprocess.DEVNULL,
        stderr=asyncio.subprocess.DEVNULL,
        cwd=cwd,
    )
    headExists_result = await headExists_proc.wait()
    headExists = headExists_result == 0
    hasNoChangesCmd = getHasNoChangesCmd(headExists)

    checks = [
        runAsync(
            *hasNoChangesCmd,
            stdout=asyncio.subprocess.DEVNULL,
            stderr=asyncio.subprocess.DEVNULL,
            cwd=cwd,
        ),
        runAsync(
            *allFilesAreTrackedCmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
            cwd=cwd,
        ),
    ]

    noChanges_proc, allTracked_proc = await resolveAll(checks)
    results = await resolveAll(
        [noChanges_proc.wait(), allTracked_proc.communicate()]
    )
    noChanges_result, allTracked_result = results

    if allTracked_proc.returncode != 0 or noChanges_result != 0:
        return False

    stdout, _stderr = allTracked_result
    return len(stdout.decode("utf-8")) == 0


def checkSync(cwd=None):
    cwd = cleanCwd(cwd)

    isAGitRepo = (
        run(
            isAGitRepoCmd,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            cwd=cwd,
        ).returncode
        == 0
    )

    if not isAGitRepo:
        raise Exception(f"'{cwd} is not a git repository")

    headExists = (
        run(
            headExistsCmd,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            cwd=cwd,
        ).returncode
        == 0
    )

    hasNoStagedChangesCmd = getHasNoChangesCmd(headExists, staged=True)
    hasNoUnstagedChangesCmd = getHasNoChangesCmd(headExists)

    allTracked_result = run(
        allFilesAreTrackedCmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
        cwd=cwd,
    )

    #
    # This is required in order to get consistent results with git-diff-index
    #
    # for more info
    # https://stackoverflow.com/questions/34807971/why-does-git-diff-index-head-result-change-for-touched-files-after-git-diff-or-g
    #

    run(
        refreshIndexCmd,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        cwd=cwd,
    )

    hasNoStagedChanges_result = run(
        hasNoStagedChangesCmd,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        cwd=cwd,
    )

    hasNoUnstagedChanges_result = run(
        hasNoUnstagedChangesCmd,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        cwd=cwd,
    )

    return (
        hasNoStagedChanges_result.returncode == 0
        and hasNoUnstagedChanges_result.returncode == 0
        and allTracked_result.returncode == 0
        and len(allTracked_result.stdout) == 0
    )


# ------- #
# Helpers #
# ------- #


def cleanCwd(cwd):
    if cwd is None:
        return os.getcwd()
    elif not isinstance(cwd, str):
        raise TypeError(
            dedent(
                f"""\
                cwd must be an instance of str
                type(cwd): {type(cwd)}
                cwd: {cwd}
                """
            )
        )

    return cwd


def getHasNoChangesCmd(headExists, staged=False):
    cmd = [
        "git",
        "diff-index",
        "--quiet",
        iif(headExists, "HEAD", emptyGitTree),
        "--",
    ]

    if staged:
        cmd.insert(2, "--cached")

    return cmd



# ------- #
# Imports #
# ------- #

from .meta import version
from traceback import format_exc
from textwrap import dedent
from types import SimpleNamespace as o
from .check import checkSync
from .utils import iif
import os


# ---- #
# Init #
# ---- #

arguments = set(["--dir", "--silent"])
helpOrVersion = set(["--help", "--version"])

twoLineSeps = os.linesep + os.linesep

usage = dedent(
    f"""
    Usage
      is-git-repo-clean [options]
      is-git-repo-clean (--help | --version)

    Options
      --dir:      path to the git repo to test.  Defaults to `os.getcwd()`
      --silent:   a flag which disables output

    Returns
      <exit code>: <output>

      0: yes
      1: no
      2: <invalid arg message>
      3: dir is not a git repository
      4: unexpected error occurred <error>
    """
)


# ---- #
# Main #
# ---- #


def getIsGitRepoClean(args):
    result = o(stdout=None, stderr=None, code=None)

    numArgs = len(args)
    if numArgs == 1:
        if args[0] == "--help":
            result.stdout = usage
            result.code = 0
            return result
        elif args[0] == "--version":
            result.stdout = version
            result.code = 0
            return result

    argsObj = validateAndParseArgs(args, result)

    if argsObj is result:
        return result

    isSilent = argsObj.silent

    try:
        isClean = checkSync(argsObj.dir or None)
        if not isSilent:
            if isClean:
                result.stdout = "yes"
            else:
                result.stderr = "no"

        result.code = iif(isClean, 0, 1)
        return result

    except Exception as e:
        if "is not a git repository" in str(e):
            if not isSilent:
                result.stderr = "dir is not a git repository"

            result.code = 3
        else:
            if not isSilent:
                result.stderr = (
                    f"unexpected error occurred{twoLineSeps}" + format_exc()
                )

            result.code = 4

        return result


# ------- #
# Helpers #
# ------- #


def validateAndParseArgs(args, result):
    i = 0
    argsObj = o(dir=None, silent=False)

    while i < len(args):
        arg = args[i]
        if arg not in arguments:
            if arg in helpOrVersion:
                result.stderr = f"'{arg}' must be the only argument when passed"
            else:
                result.stderr = f"invalid option '{arg}'"

            result.stderr += os.linesep + usage
            result.code = 2

            return result

        if arg == "--dir":
            if i == len(args) - 1:
                result.stderr = "'--dir' must be given a value"
                result.stderr += os.linesep + usage
                result.code = 2

                return result

            argsObj.dir = args[i + 1]
            i += 1
        else:
            argsObj.silent = True

        i += 1

    return argsObj



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['is_git_repo_clean']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['is-git-repo-clean = is_git_repo_clean.script:main']}

setup_kwargs = {
    'name': 'is-git-repo-clean',
    'version': '0.3.5',
    'description': 'A simple function to test whether your git repo is clean',
    'long_description': "## Is Git Repo Clean\n\n<!-- START doctoc generated TOC please keep comment here to allow auto update -->\n<!-- DON'T EDIT THIS SECTION, INSTEAD RE-RUN doctoc TO UPDATE -->\n**Table of Contents**\n\n- [Intro](#intro)\n    - [What is it?](#what-is-it)\n    - [Why create it?](#why-create-it)\n- [Install](#install)\n- [Usage](#usage)\n    - [programmatic](#programmatic)\n    - [cli](#cli)\n- [Api](#api)\n    - [`check` async (dir=os.getcwd()) => bool](#check-async-dirosgetcwd--bool)\n    - [`checkSync` (dir=os.getcwd()) => bool](#checksync-dirosgetcwd--bool)\n    - [`version`](#version)\n- [Test](#test)\n\n<!-- END doctoc generated TOC please keep comment here to allow auto update -->\n\n<br>\n\n### Intro\n\n#### What is it?\n\nA simple function that tests whether your git repo is clean.\n\nby *clean* I mean it has:\n- no untracked files\n- no staged changes\n- no unstaged changes\n\nInstalling this also exposes a cli command `is-git-repo-clean`\n\n\n#### Why create it?\n\nI wanted to write a build script that would exit early if the git repo\nwasn't clean\n\n<br>\n\n### Install\n\n```sh\n$ pip install is_git_repo_clean\n```\n\n<br>\n\n### Usage\n\n#### programmatic\n\n```python\nimport is_git_repo_clean\n\n\nasync def isCleanAsync(pathToGitRepo = None):\n  # async by default\n   return await is_git_repo_clean.check(pathToGitRepo)\n\n\ndef isCleanSync(pathToGitRepo = None):\n  # sync available\n  return is_git_repo_clean.checkSync(pathToGitRepo)\n```\n\n#### cli\n\n```sh\n$ is-git-repo-clean --help\n\nUsage\n  is-git-repo-clean [--dir <path>] [--silent]\n  is-git-repo-clean --help\n  is-git-repo-clean --version\n\nOptions\n  dir:      path to the git repo to test.  Defaults to `os.getcwd()`\n  silent:   disables output\n  help:     print this\n  version:  prints the version of this tool\n\nReturns\n  <exit code>: <output>\n\n  0: yes\n  1: no\n  2: <invalid arg message>\n  3: dir is not a git repository\n  4: unexpected error occurred <error>\n```\n\n<br>\n\n### Api\n\n`is_git_repo_clean` exports the following\n\n\n#### `check` async (dir=os.getcwd()) => bool\n - an asynchronous function that returns whether the git repo is clean\n\n\n#### `checkSync` (dir=os.getcwd()) => bool\n - a synchronous function that returns whether the git repo is clean\n\n\n#### `version`\n\n<br>\n\n### Test\n\n```sh\nhub clone olsonpm/py_is-git-repo-clean\ncd py_is-git-repo-clean\npython runTests.py\n```\n",
    'author': 'Philip Olson',
    'author_email': 'philip.olson@pm.me',
    'url': 'https://github.com/olsonpm/py_is-git-repo-clean',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
